# Copyright 2018 Mario Graff

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_space():
    from microtc.textmodel import TextModel
    from microtc.weighting import TFIDF
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    docs = [x['text'] for x in tw]
    text = TextModel(docs, token_list=[-1, 3])
    # print(text['buenos dias'])
    docs = [text.tokenize(d) for d in docs]
    sp = TFIDF(docs)
    assert len(sp.wordWeight) == len(sp._w2id)
    # print(sp._weight)
    # print(sp._w2id)


def test_doc2weight():
    from microtc.textmodel import TextModel
    from microtc.weighting import TFIDF
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    docs = [x['text'] for x in tw]
    text = TextModel(docs, token_list=[-1, 3])
    # print(text['buenos dias'])
    docs = [text.tokenize(d) for d in docs]
    sp = TFIDF(docs)
    assert len(sp.doc2weight(text.tokenize('odio odio los los'))) == 3


def test_getitem():
    from microtc.textmodel import TextModel
    from microtc.weighting import TFIDF
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    docs = [x['text'] for x in tw]
    text = TextModel(docs, token_list=[-1, 3])
    # print(text['buenos dias'])
    docs = [text.tokenize(d) for d in docs]
    sp = TFIDF(docs)
    tok = text.tokenize('buenos dias')
    bow = sp.doc2weight(tok)
    ids, tf, df = bow
    assert len(ids) == len(sp[tok])


def test_entropy():
    from microtc.textmodel import TextModel
    from microtc.weighting import Entropy, TFIDF
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    docs = [x['text'] for x in tw]
    text = TextModel(token_list=[-1, 3])
    # print(text['buenos dias'])
    docs = [text.tokenize(d) for d in docs]
    sp = Entropy(docs, X=tw)
    print(sp.wordWeight)
    tfidf = TFIDF(docs)
    for k in sp.wordWeight.keys():
        if sp.wordWeight[k] != tfidf.wordWeight[k]:
            return
    print(sp.w)
    assert False


    



# Copyright 2018 Mario Graff

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from collections import Counter
import os


TEXT = os.environ.get("TEXT", 'text')
KLASS = os.environ.get("KLASS", 'klass')
VALUE = os.environ.get("VALUE", 'value')


class TFIDF(object):
    """
    Vector Space model using TFIDF

    :param docs: corpus as a list of list of tokens
    :type docs: list
    :param X: original corpus, useful to pass extra information in a dict
    :type X: list
    :param token_min_filter: Keep those tokens that appear more times than the parameter
    :type token_min_filter: int or float

    :param token_max_filter: Keep those tokens that appear less times than the parameter
    :type token_max_filter: int or float

    Usage:

    >>> from microtc.weighting import TFIDF
    >>> tokens = [['buenos', 'dia', 'microtc'], ['excelente', 'dia'], ['buenas', 'tardes'], ['las', 'vacas', 'me', 'deprimen'], ['odio', 'los', 'lunes'], ['odio', 'el', 'trafico'], ['la', 'computadora'], ['la', 'mesa'], ['la', 'ventana']]
    >>> tfidf = TFIDF(tokens)
    >>> vector = tfidf['buenos', 'X', 'trafico']
    """

    def __init__(self, docs, X=None, token_min_filter=0, token_max_filter=1):
        w2id = {}
        weight = {}
        self._ndocs = len(docs)
        for tokens in docs:
            for x, freq in Counter(tokens).items():
                try:
                    ident = w2id[x]
                    weight[ident] = weight[ident] + 1
                except KeyError:
                    ident = len(w2id)
                    w2id[x] = ident
                    weight[ident] = 1
        if token_min_filter > 0 or token_max_filter != 1:
            if token_min_filter < 1:
                token_min_filter = int(self._ndocs * token_min_filter)
                if token_min_filter < 1:
                    token_min_filter = 1
            if token_min_filter > 0:
                w2id = [(k, v) for k, v in w2id.items() if weight[v] > token_min_filter]
            if token_max_filter != 1:
                if token_max_filter < 1:
                    token_max_filter = int(self._ndocs * token_max_filter)
                w2id = [(k, v) for k, v in w2id if weight[v] < token_max_filter]
            w2id.sort(key=lambda x: x[1])
            mm = {k: v[1] for k, v in enumerate(w2id)}
            w2id = {v[0]: k for k, v in enumerate(w2id)}
            weight = {ident: weight[mm[ident]] for ident in w2id.values()}
        self.word2id = w2id
        self.wordWeight = weight

    @property
    def num_terms(self):
        """Number of terms"""

        return self._num_terms

    @property
    def word2id(self):
        """Map word to id"""

        return self._w2id

    @word2id.setter
    def word2id(self, value):
        self._num_terms = len(value)
        self._w2id = value

    @property
    def wordWeight(self):
        """Weight associated to each word, this could be the inverse document frequency"""
        return self._weight

    @wordWeight.setter
    def wordWeight(self, value):
        """Inverse document frequency

        :param value: weights
        :type value: dict
        """

        N = self._ndocs
        self._weight = {k: np.log2(N / v) for k, v in value.items()}

    def doc2weight(self, tokens):
        """Weight associated to each token

        :param tokens: list of tokens
        :type tokens: lst

        :rtype: tuple - ids, term frequency, wordWeight
        """
        lst = []
        w2id = self._w2id
        weight = self.wordWeight
        for token in tokens:
            try:
                id = w2id[token]
                lst.append(id)
            except KeyError:
                continue
        ids_tf = [(a, b) for a, b in Counter(lst).items()]
        # ids, tf = np.unique(lst, return_counts=True)
        ids = [x[0] for x in ids_tf]
        tf = np.array([x[1] for x in ids_tf])
        tf = tf / tf.sum()
        df = np.array([weight[x] for x in ids])
        return ids, tf, df

    def __getitem__(self, tokens):
        """
        TF-IDF and the vectors are normalised.

        :param tokens: list of tokens
        :type tokens: lst

        :rtype: lst
        """

        __ = self.doc2weight(tokens)
        r = [(i, _tf * _df) for i, _tf, _df in zip(*__)]
        n = np.sqrt(sum([x * x for _, x in r]))
        return [(i, x/n) for i, x in r]


class TF(TFIDF):
    @property
    def wordWeight(self):
        """Weight associated to each word, this is one on TF"""
        return self._weight

    @wordWeight.setter
    def wordWeight(self, value):
        """Inverse document frequency

        :param value: weights
        :type value: dict
        """

        self._weight = {k: 1 for k, v in value.items()}

    def __getitem__(self, tokens):
        """
        TF, the frequency is normalised

        :param tokens: list of tokens
        :type tokens: lst

        :rtype: lst
        """

        __ = self.doc2weight(tokens)
        r = [(i, _tf) for i, _tf, _df in zip(*__)]
        return r


class Entropy(TF):
    """
    Vector Space using 1 - entropy as the weighting scheme

    Usage:

    >>> from microtc.weighting import Entropy
    >>> tokens = [['buenos', 'dia', 'microtc'], ['excelente', 'dia'], ['buenas', 'tardes'], ['las', 'vacas', 'me', 'deprimen', 'al', 'dia'], ['odio', 'los', 'lunes'], ['odio', 'el', 'trafico'], ['la', 'computadora'], ['la', 'mesa'], ['la', 'ventana']]
    >>> y = [0, 0, 0, 2, 2, 2, 1, 1, 1]
    >>> ent = Entropy(tokens, X=[dict(text=t, klass=k) for t, k in zip(tokens, y)])
    >>> vector = ent['buenos', 'X', 'dia']
    """
    def __init__(self, docs, X=None, **kwargs):
        assert X is not None
        super(Entropy, self).__init__(docs, X=X, **kwargs)
        self.wordWeight = self.entropy(docs, X, self.word2id)

    @property
    def wordWeight(self):
        """Weight associated to each word, entropy per token"""
        return self._weight

    @wordWeight.setter
    def wordWeight(self, value):
        """Entropy

        :param value: weights
        :type value: dict
        """

        if isinstance(value, dict):
            self._weight = value
        else:
            self._weight = {k: v for k, v in enumerate(value)}

    @staticmethod
    def entropy(corpus, docs, word2id):
        """
        Compute entropy

        :param corpus: Tokenized corpus, i.e., as a list of tokens list
        :type corpus: list
        :param docs: Original corpus is a list of dictionaries where key klass contains the class or label
        :type docs: list
        :param word2id: Map token to identifier
        :type word2id: dict

        :rtype: np.array
        """
        m = word2id
        y = [x[KLASS] for x in docs]
        klasses = np.unique(y)
        nklasses = klasses.shape[0]
        ntokens = len(m)
        # hist = np.ones((klasses.shape[0], ntokens))
        hist = np.full((klasses.shape[0], ntokens), 3)

        for ki, klass in enumerate(klasses):
            for _y, tokens in zip(y, corpus):
                if _y != klass:
                    continue
                for x in Counter(tokens).keys():
                    try:
                        hist[ki, m[x]] += 1
                    except KeyError:
                        continue

        # hist = np.log2(hist + 1)
        hist = hist / hist.sum(axis=0)
        # hist[~np.isfinite(hist)] = 1.0 / nklasses
        logc = np.log2(hist)
        logc[~np.isfinite(logc)] = 0
        if nklasses > 2:
            logc = logc / np.log2(nklasses)
        return (1 + (hist * logc).sum(axis=0))

    def __getitem__(self, tokens):
        """
        Entropy

        :param tokens: list of tokens
        :type tokens: lst

        :rtype: lst
        """

        __ = self.doc2weight(tokens)
        r = [(i, _df) for i, _tf, _df in zip(*__)]
        return r



# author: Eric S. Tellez


def test_params():
    from microtc.params import ParameterSelection
    import numpy as np
    from numpy.random import random
    sel = ParameterSelection()
    
    def fake_score(conf_code):
        conf = conf_code[0]
        conf['_score'] = random()
        conf['_time'] = 1.0
        return conf
        
    sel.search(fake_score, bsize=64)


def test_read_data_labels():
    import os
    from microtc.utils import read_data_labels
    filename = os.path.join(os.path.dirname(__file__), "text.json")
    read_data_labels(filename)


def test_wrapper_score():
    from microtc.scorewrapper import ScoreKFoldWrapper
    from sklearn.metrics import f1_score
    import numpy as np
    np.random.seed(0)
    y = np.random.randint(3, size=100).astype(np.str)
    hy = np.random.randint(3, size=100)
    w = ScoreKFoldWrapper([], y, score='avgf1:0:2', nfolds=10)
    conf = {}
    w.compute_score(conf, hy)
    f1 = f1_score(y.astype(np.int), hy, average=None)
    assert conf['_accuracy'] == (y.astype(np.int) == hy).mean()
    print(y)
    print(conf['_avgf1:0:2'], (f1[0] + f1[2]) / 2.)
    assert conf['_avgf1:0:2'] == (f1[0] + f1[2]) / 2.


def test_counter():
    from microtc.utils import Counter, save_model, load_model
    import os
    c = Counter()
    c.update([1, 2, 3, 1])
    c.update([3])
    assert c[1] == 2
    print(c.update_calls)
    assert c.update_calls == 2
    save_model(c, "t.voc")
    cc = load_model("t.voc")
    os.unlink("t.voc")
    print(cc.update_calls, "**")
    assert cc.update_calls ==  2


# Copyright 2016 Mario Graff (https://github.com/mgraffg) and Ranyart R. Suarez (https://github.com/RanyartRodrigo)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import sys


def test_nparams():
    from microtc.command_line import params
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    params('-k', '2', '-s', '11', '-o', fname + ".tmp", fname)
    os.unlink(fname + ".tmp")


def test_main():
    from microtc.command_line import params
    import os
    import tempfile
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    params('-o', output, '-k', '2', fname)
    os.unlink(output)


def test_pool():
    from microtc.command_line import params
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    params('-k', '2', '-s', '11', '-n', '2', '-o', fname + ".params", fname)
    os.unlink(fname + ".params")


def test_output():
    from microtc.command_line import params
    import os
    import tempfile
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    params('-o', output, '-k', '2', fname)
    assert os.path.isfile(output)
    os.unlink(output)


def test_seed():
    try:
        from mock import MagicMock
    except ImportError:
        from unittest.mock import MagicMock
    from microtc.command_line import params
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    seed = np.random.seed
    np.random.seed = MagicMock()
    params('-s', '2', '--seed', '1', '-k', '2', '-o', fname + ".params", fname)
    os.unlink(fname + ".params")
    np.random.seed.assert_called_once_with(1)
    np.random.seed = seed


def test_train():
    from microtc.command_line import params, train
    import os
    import tempfile
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    sys.argv[0] = 'test-train:microtc-params'
    params('-o', output, '-k', '2', fname, '-s', '2')
    assert os.path.isfile(output)
    with open(output) as fpt:
        print(fpt.read())

    sys.argv[0] = 'test-train:microtc-train'
    print(train('-m', output, '-o', output + ".model", fname))
    os.unlink(output)
    os.unlink(output + '.model')


def test_train_regression():
    from microtc.command_line import params, train, predict
    import os
    import sys
    import tempfile

    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    params('-o', output, '-k', '2', '-s', '2', '-S', 'r2', fname)
    with open(output) as fpt:
        print("XXXXX>", output, fpt.read())

    train('-o', output + '.model', '-R', '-m', output, fname)
    for x in predict('-m', output + '.model', fname, '-o', output + '.predicted'):
        print(x, file=sys.stderr)

    os.unlink(output)
    os.unlink(output + '.model')
    os.unlink(output + '.predicted')


def test_train2():
    from microtc.command_line import train, params
    import os
    import tempfile
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    params('-o', output, '-k', '2', fname, '-s', '2')
    assert os.path.isfile(output)
    output2 = tempfile.mktemp()
    train('-m', output, fname, '-o', output2)
    os.unlink(output)
    os.unlink(output2)


def test_test():
    from microtc.command_line import params, train, predict
    from microtc.utils import read_data_labels
    import os
    import sys
    import tempfile
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    params('-o', output, '-k', '0.5:0.5', fname, '-s', '2')
    params('-o', output, '-k', '2', fname, '-s', '2')
    train('-o', output + '.model', '-m', output, fname)
    predict('-m', output + '.model', fname, '-o', output + '.predicted')
    X, y = read_data_labels(fname)
    print(y)
    os.unlink(output)
    os.unlink(output + '.model')
    os.unlink(output + '.predicted')
    assert len(y)


def test_score():
    from microtc.command_line import params
    import os
    import sys
    import tempfile
    import json
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    params('-o', output, '-k', '2', '-s', '2', '-S', 'avgf1:POS:NEG', fname)
    with open(output) as fpt:
        a = json.loads(fpt.read())[0]
    assert a['_score'] == a['_avgf1:POS:NEG']
    os.unlink(output)
        

def test_textmodel():
    from microtc.command_line import params, train, textmodel
    import os
    import sys
    import json
    import tempfile
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    params('-o', output, '-k', '2', fname, '-s', '2')
    train('-m', output, fname, '-o', output)
    output2 = tempfile.mktemp()
    textmodel('-m', output, fname, '-o', output2)
    os.unlink(output)
    a = open(output2).readline()
    os.unlink(output2)
    a = json.loads(a)
    assert 'klass' in a


def test_numeric_klass():
    from microtc.utils import tweet_iterator
    from microtc.params import DefaultParams, Fixed
    from microtc.command_line import params, train, predict
    from sklearn.preprocessing import LabelEncoder
    import os
    import json
    import tempfile
    import sys
    numeric = tempfile.mktemp() + '.json'
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    D = [x for x in tweet_iterator(fname)]
    encoder = LabelEncoder().fit([x['klass'] for x in D])
    y = encoder.transform([x['klass'] for x in D])
    for x, k in zip(D, y):
        x['klass'] = int(k)

    with open(numeric, 'w') as fpt:
        [fpt.write(json.dumps(x) + '\n') for x in D]

    P = DefaultParams.copy()
    params('-o', output, '-k', '2', numeric, '-s', '2', **P)
    train('-m', output, numeric, '-o', output)
    output2 = tempfile.mktemp()
    predict('-m', output, fname, '-o', output2)
    os.unlink(numeric)
    os.unlink(output)
    os.unlink(output2)


def test_kfolds():
    from microtc.command_line import params, kfolds
    import os
    import sys
    import json
    import tempfile
    output = tempfile.mktemp()
    fname = os.path.dirname(__file__) + '/text.json'
    params('-o', output, '-k', '2', fname, '-s', '2')
    output2 = tempfile.mktemp()
    kfolds('-m', output, fname, '-o', output2)
    os.unlink(output)
    a = open(output2).readline()
    os.unlink(output2)
    a = json.loads(a)
    assert 'decision_function' in a
    try:
        kfolds('--update-klass', '-m', output, fname, '-o', output2)
    except AssertionError:
        return
    assert False



# Copyright 2016 Mario Graff (https://github.com/mgraffg)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_tweet_iterator():
    import os
    import gzip
    from microtc.utils import tweet_iterator
    
    fname = os.path.dirname(__file__) + '/text.json'
    a = [x for x in tweet_iterator(fname)]
    fname_gz = fname + '.gz'
    with open(fname, 'r') as fpt:
        with gzip.open(fname_gz, 'w') as fpt2:
            fpt2.write(fpt.read().encode('ascii'))
    b = [x for x in tweet_iterator(fname_gz)]
    assert len(a) == len(b)
    for a0, b0 in zip(a, b):
        assert a0['text'] == b0['text']
    os.unlink(fname_gz)


def test_textmodel():
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    text = TextModel([x['text'] for x in tw])
    # print(text.tokenize("hola amiguitos gracias por venir :) http://hello.com @chanfle"))
    # assert False
    assert isinstance(text[tw[0]['text']], list)
    assert len(text[tw[0]]) == 3
    text = TextModel(token_list=[3]).fit(tw)
    print(text.model.word2id)
    for k, _ in text.model.word2id.items():
        assert len(k) == 5


def test_params():
    import os
    import itertools
    from microtc.params import BASIC_OPTIONS
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator

    params = dict(del_diac=[True, False], usr_option=BASIC_OPTIONS,
                  url_option=BASIC_OPTIONS)
    params = sorted(params.items())
    fname = os.path.dirname(__file__) + '/text.json'
    tw = [x for x in tweet_iterator(fname)]
    text = [x['text'] for x in tw]
    for x in itertools.product(*[x[1] for x in params]):
        args = dict(zip([x[0] for x in params], x))
        ins = TextModel(text, **args)
        assert isinstance(ins[text[0]], list)


def test_lang():
    from microtc.textmodel import TextModel

    text = [
        "Hi :) :P XD",
        "excelente dia xc",
        "el alma de la fiesta XD"
    ]
    model = TextModel(text, **{
        "del_dup": True,
        "emo_option": "group",
        "lc": True,
        "num_option": "group",
        "del_diac": False,
        "token_list": [
            (2, 1),
            (2, 2),
            -1,
            # 5,
        ],
        "url_option": "group",
        "usr_option": "group",
    })
    text = "El alma de la fiesta :) conociendo la maquinaria @user bebiendo nunca manches que onda"
    a = model.tokenize(text)
    b = ['el~de', 'alma~la', 'de~fiesta', 'la~_pos', 'fiesta~conociendo', '_pos~la', 'conociendo~maquinaria', 'la~_usr', 'maquinaria~bebiendo', '_usr~nunca',
         'bebiendo~manches', 'nunca~que', 'manches~onda', 'el~la', 'alma~fiesta', 'de~_pos', 'la~conociendo', 'fiesta~la', '_pos~maquinaria', 'conociendo~_usr',
         'la~bebiendo', 'maquinaria~nunca', '_usr~manches', 'bebiendo~que', 'nunca~onda', 'el', 'alma', 'de', 'la', 'fiesta', '_pos',
         'conociendo', 'la', 'maquinaria', '_usr', 'bebiendo', 'nunca', 'manches', 'que', 'onda']
    print(text)
    assert a == b, "got: {0}, expected: {1}".format(a, b)


def test_textmodel_token_min_filter():
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    text = TextModel(tw, token_min_filter=1, token_list=[-2, -1, 3, 4])
    print(len(text.model._w2id), 'hh', text.token_min_filter, text.token_max_filter,
          text.model._w2id)
    assert len(text.model._w2id) == 30
    text = TextModel(tw, token_min_filter=0.01, token_list=[-2, -1, 3, 4])
    print(len(text.model._w2id))
    assert len(text.model._w2id) == 30
    text = TextModel(tw, token_min_filter=1)


def test_textmodel_token_max_filter():
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    text = TextModel(tw, token_max_filter=len(tw) / 2, token_list=[-2, -1, 3, 4])
    print(len(text.model._w2id))
    assert len(text.model._w2id) == 30
    text = TextModel(tw, token_max_filter=0.5, token_list=[-2, -1, 3, 4])
    print(len(text.model._w2id))
    assert len(text.model._w2id) == 29
    text = TextModel(tw, token_max_filter=2)
    print(len(text.model._w2id))


def test_textmodel_entropy():
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    text = TextModel(tw, weighting='microtc.weighting.Entropy', token_list=[-1, 3])
    # print(text.tokenize("hola amiguitos gracias por venir :) http://hello.com @chanfle"))
    # assert False
    assert isinstance(text[tw[0]['text']], list)
    _ = text[tw[0]]
    print(_)
    for k, v in _:
        assert text.model.wordWeight[k] == v


def test_textmodel_transform_tonp():
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator
    from sklearn.svm import LinearSVC
    from sklearn.preprocessing import LabelEncoder
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    text = TextModel().fit(tw)
    X = text.transform(tw)
    le = LabelEncoder().fit([x['klass'] for x in tw])
    y = le.transform([x['klass'] for x in tw])
    m = LinearSVC().fit(text.tonp(X), y)
    assert len(m.predict(text.tonp(X))) == len(y)


def test_textmodel_compute_tokens():
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    tm = TextModel(token_list=[-2, -1])
    text = tm.text_transformations(tw[0]['text'])
    L = tm.compute_tokens(text)
    assert len(L) == 2
    r = []
    [r.__iadd__(x) for x in L]
    for a, b in zip(tm.tokenize(tw[0]), r):
        assert a == b


def test_textmodel_weighting_key():
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    for w in ['tfidf', 'tf', 'entropy']:
        TextModel(token_list=[-2, -1], weighting=w).fit(tw)


def test_textmodel_save_load():
    import os
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator, save_model, load_model
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    tm = TextModel().fit(tw)
    save_model(tm, 't.model')
    assert isinstance(load_model('t.model'), TextModel)
    os.unlink('t.model')
    
    
def test_textmodel_num_terms():
    from microtc.textmodel import TextModel
    from microtc.utils import tweet_iterator
    import os
    fname = os.path.dirname(__file__) + '/text.json'
    tw = list(tweet_iterator(fname))
    text = TextModel(token_list=[-2, -1, 3, 4]).fit(tw)
    assert text.num_terms is not None
    text.transform(["buenos"])
    print(text.num_terms)
    assert text.num_terms == text.model.num_terms



# author: Eric S. Tellez <eric.tellez@infotec.mx>

import os
import sys
import json
import numpy as np
from itertools import combinations


try:
    from tqdm import tqdm
except ImportError:
    def tqdm(x, **kwargs):
        return x


class Fixed:
    def __init__(self, value):
        self.value = value
        self.valid_values = [value]

    def neighborhood(self, v):
        return []

    def get_random(self):
        return self.value


class SetVariable:
    def __init__(self, values):
        self.valid_values = list(values)

    def neighborhood(self, value):
        return [u for u in self.valid_values if u != value]

    def get_random(self):
        i = np.random.randint(len(self.valid_values))
        return self.valid_values[i]


class PowersetVariable:
    def __init__(self, initial_set, max_size=None):
        self.valid_values = []
        if max_size is None:
            max_size = len(initial_set) // 2 + 1

        for i in range(1, len(initial_set)+1):
            for l in combinations(initial_set, i):
                if len(l) <= max_size:
                    self.valid_values.append(l)

    def mismatches(self, value):
        lvalue = len(value)
        for v in self.valid_values:
            # if len(value.intersection(v)) == lvalue - 1 or len(value.union(v)) == lvalue + 1:
            ulen = len(value.union(v))
            ilen = len(value.intersection(v))
            if ulen in (lvalue, lvalue + 1) and ilen in (lvalue, lvalue - 1):
                yield v

    def neighborhood(self, value):
        L = []
        for v in value:
            if isinstance(v, list):
                v = tuple(v)
            L.append(v)

        return list(self.mismatches(set(L)))

    def get_random(self):
        x = len(self.valid_values)
        i = np.random.randint(x)
        return self.valid_values[i]


OPTION_NONE = 'none'
OPTION_GROUP = 'group'
OPTION_DELETE = 'delete'
BASIC_OPTIONS = [OPTION_DELETE, OPTION_GROUP, OPTION_NONE]


def Option():
    return SetVariable(BASIC_OPTIONS)


def Uniform(left, right, k=10):
    d = (right - left) * np.random.random_sample(k) + left
    return SetVariable(d)


def Normal(mean, sigma, k=10):
    d = mean * np.random.randn(k) + sigma
    return SetVariable(d)


def Boolean():
    return SetVariable([False, True])


TOKENLIST = [(3, 1), (2, 2), (2, 1), -3, -2, -1, 1, 2, 3, 5, 7, 9]
if "TOKENLIST" in os.environ:
    def _simple_cast(x):
        if isinstance(x, list):
            return tuple(x)
        else:
            return x
    
    TOKENLIST = [_simple_cast(x) for x in json.loads(os.environ["TOKENLIST"])]

MAXTOKENLIST = os.environ.get("MAXTOKENLIST", len(TOKENLIST)//2 + 1)


DefaultParams = dict(
    num_option=Option(),
    usr_option=Option(),
    url_option=Option(),
    emo_option=Option(),

    ent_option=Fixed(OPTION_NONE),
    # hashtag_option=Fixed(OPTION_NONE),
    hashtag_option=Option(),

    select_ent=Fixed(False),
    select_suff=Fixed(False),
    select_conn=Fixed(False),
    
    lc=Boolean(),
    del_dup=Boolean(),
    del_punc=Boolean(),
    del_diac=Boolean(),
    
    token_list=PowersetVariable(TOKENLIST, max_size=MAXTOKENLIST),
    # negative values means for absolute frequencies, positive values between 0 and 1 means for ratio
    token_min_filter=Fixed(-1),
    token_max_filter=Fixed(1.0),
    # token_max_filter=SetVariable([0.5, 0.9, 0.95, 0.99, 1.0]),
    # token_min_filter=SetVariable([-1, -2, -3, -5, -7, -9]),
    weighting=SetVariable(['tfidf', 'tf', 'entropy']),
)

if "PARAMS" in os.environ:
    for k, v in json.loads(os.environ["PARAMS"]).items():
        DefaultParams[k] = Fixed(v)


class ParameterSelection:
    MINIMUM_IMPROVEMENT = 0.001
    IMPROVEMENT_FAILURES = 2

    def __init__(self, params=None):
        if (params is None) or (0 == len(params)):
            params = DefaultParams
        else:
            for k in DefaultParams.keys():
                assert k in params, "{0} is not in given parameters; {1}".format(k, params)

        self.params = params

    def sample_param_space(self, n):
        for i in range(n):
            kwargs = {}
            for k, v in self.params.items():
                kwargs[k] = v.get_random()

            yield kwargs

    def expand_neighbors(self, s, keywords=None):
        if keywords is None:
            keywords = set(s.keys())

        for k, v in sorted(s.items()):
            if k[0] == '_' or k not in keywords:
                # by convention, metadata starts with underscore
                continue

            vtype = self.params[k]
            if isinstance(vtype, Fixed):
                continue

            for neighbor in vtype.neighborhood(v):
                x = s.copy()
                x[k] = neighbor
                yield(x)

    def get_best(self, fun_score, cand, desc="searching for params", pool=None):
        if pool is None:
            # X = list(map(fun_score, cand))
            X = [fun_score(x) for x in tqdm(cand, desc=desc, total=len(cand))]
        else:
            # X = list(pool.map(fun_score, cand))
            X = [x for x in tqdm(pool.imap_unordered(fun_score, cand), desc=desc, total=len(cand))]

        # a list of tuples (score, conf)
        X.sort(key=lambda x: x['_score'], reverse=True)
        return X

    def search(self, fun_score, bsize=32, hill_climbing=True, best_list=None, tabu=None, pool=None):
        if tabu is None:
            tabu = set()
        
        best_list = []
        prev = 0.0
        restarts = 0

        while restarts < ParameterSelection.IMPROVEMENT_FAILURES:
            b = self._search(fun_score, bsize=bsize, hill_climbing=hill_climbing, tabu=tabu, pool=pool)
            best_list.extend(b)
            best_list.sort(key=lambda x: x['_score'], reverse=True)
            curr = best_list[0]['_score']
            if curr - prev < ParameterSelection.MINIMUM_IMPROVEMENT:
                restarts += 1
            
            prev = curr

            print("*** best configuration found (restart failures: {0} of {1})".format(restarts, ParameterSelection.IMPROVEMENT_FAILURES), file=sys.stderr)
            print(json.dumps(best_list[0], sort_keys=True), file=sys.stderr)

        # best_list.sort(key=lambda x: x['_score'], reverse=True)
        return best_list
    
    def _search(self, fun_score, bsize=32, hill_climbing=True, best_list=None, tabu=None, pool=None):
        # initial approximation, montecarlo based procesess

        if tabu is None:
            tabu = set()  # memory for tabu search

        if best_list is None:
            L = []
            for conf in self.sample_param_space(bsize):
                code = get_filename(conf)
                if code in tabu:
                    continue

                tabu.add(code)
                L.append((conf, code))
            
            best_list = self.get_best(fun_score, L, pool=pool)
        else:
            for conf in best_list:
                tabu.add(get_filename(conf))

        def _hill_climbing(keywords, desc):
            # second approximation, a hill climbing process
            i = 0
            while True:
                i += 1
                prev = best_list[0]['_score']
                L = []

                for conf in self.expand_neighbors(best_list[0], keywords=keywords):
                    code = get_filename(conf)
                    if code in tabu:
                        continue

                    tabu.add(code)
                    L.append((conf, code))

                if len(L) > bsize:
                    print("Selecting {0} random configuration from a neighborhood of {1}".format(bsize, len(L)), file=sys.stderr)
                    np.random.shuffle(L)
                    L = L[:bsize]
        
                best_list.extend(self.get_best(fun_score, L, desc=desc + " {0}".format(i), pool=pool))
                best_list.sort(key=lambda x: x['_score'], reverse=True)
                curr = best_list[0]['_score']
                if curr - prev < ParameterSelection.MINIMUM_IMPROVEMENT:
                    break

        if hill_climbing:
            print("starting hill climbing with configuration", file=sys.stderr)
            print(json.dumps(best_list[0], sort_keys=True), file=sys.stderr)
            _hill_climbing(['token_list'], "optimizing token_list")
            # _hill_climbing(['token_min_filter', 'token_max_filter'], "optimizing token max and min filters")

            do_vectorizing_opt = len(self.params['token_min_filter'].valid_values) > 1 or len(self.params['token_max_filter'].valid_values) > 1
            if do_vectorizing_opt:
                _hill_climbing(['token_list', 'token_min_filter', 'token_max_filter'], "optimizing all token parameters")

            ks = list(self.params.keys())

            if do_vectorizing_opt:
                ks.remove('token_list')
                ks.remove('token_min_filter')
                ks.remove('token_max_filter')

            _hill_climbing(ks, "optimizing the rest of params")

        return best_list


def get_filename(kwargs, basename=None):
    L = []
    if basename:
        L.append(basename)
        
    for k, v in sorted(kwargs.items()):
        if k[0] == '_':
            continue

        L.append("{0}={1}".format(k, v).replace(" ", ""))

    return "-".join(L)



# Copyright 2020 Mario Graff

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from microtc import emoticons


def download(fname):
    from urllib import request
    import os
    # fname = "https://www.unicode.org/Public/emoji/12.1/emoji-data.txt"
    output = fname.split("/")[-1]
    if os.path.isfile(output):
        return output
    request.urlretrieve(fname, output)
    return output


def test_read_emoji_standard():
    # fname = "/home/mgraffg/software/microtc/microtc/resources/emoji-data.txt"
    data = "https://www.unicode.org/Public/emoji/12.1/emoji-data.txt"
    sec = "https://www.unicode.org/Public/emoji/12.1/emoji-sequences.txt"
    var ="https://www.unicode.org/Public/emoji/12.1/emoji-variation-sequences.txt"
    zwj = "https://www.unicode.org/Public/emoji/12.1/emoji-zwj-sequences.txt"
    emos = emoticons.read_emoji_standard(download(data))
    print(len(emos))
    emoticons.read_emoji_standard(download(sec), emos)
    print(len(emos))
    emoticons.read_emoji_standard(download(var), emos)
    print(len(emos))
    emoticons.read_emoji_standard(download(zwj), emos)
    print(chr(0x1F3FB) in emos, len(emos))
    assert (chr(0x1F468) + chr(0x200D) + chr(0x1F467)) in emos


def test_convert_emoji():
    emo = emoticons.convert_emoji("26BD")
    assert isinstance(emo, str)
    assert chr(0x26bd) == emo
    emos = emoticons.convert_emoji("2753..2755")
    assert isinstance(emos, list)
    assert len(emos) == 3
    print(emos)
    emo = emoticons.convert_emoji("1F4EA FE0E")
    print(emo)
    assert len(emo) == 2 and isinstance(emo, str)


def test_create_data_structure():
    data = "https://www.unicode.org/Public/emoji/12.1/emoji-data.txt"
    zwj = "https://www.unicode.org/Public/emoji/12.1/emoji-zwj-sequences.txt"
    emojis = emoticons.read_emoji_standard(download(data))
    emoticons.read_emoji_standard(download(zwj), emojis)
    data = emoticons.create_data_structure(dict(mario=1, ma=1, g=1, marx=1))
    print(data)
    emoticons.create_data_structure(emojis)


def test_has_emoji():
    data = emoticons.create_data_structure(dict(mario=1, ma=1, g=1, marx=1))
    text = "maxmarxgmlr"
    blocks = emoticons.find_emoji(data, text)
    lst = [text[init:end] for init, end in blocks]
    for a, b in zip(lst, ["ma", "marx", "g"]):
        assert a == b
    print(lst)
    data = "https://www.unicode.org/Public/emoji/12.1/emoji-data.txt"
    zwj = "https://www.unicode.org/Public/emoji/12.1/emoji-zwj-sequences.txt"
    emojis = emoticons.read_emoji_standard(download(data))
    emoticons.read_emoji_standard(download(zwj), emojis)
    data = emoticons.create_data_structure(emojis)
    _ = emoticons.find_emoji(data, chr(0x1F468) + chr(0x200D) + chr(0x1F468))
    print(_)
    init, end = _[0]
    assert (end - init) == 1
    assert chr(0x1F468) in emojis
    assert len(_) == 1



# Copyright 2016 Eric S. Tellez

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import json
import gzip
import collections


def line_iterator(filename):
    if filename.endswith(".gz"):
        f = gzip.GzipFile(filename)
    else:
        f = open(filename, encoding='utf8')

    while True:
        line = f.readline()
        # Test the type of the line and encode it if neccesary...
        if type(line) is bytes:
            line = str(line, encoding='utf8')

        # If the line is empty, we are done...
        if len(line) == 0:
            break

        line = line.strip()
        # If line is empty, jump to next...
        if len(line) == 0:
            continue

        yield line

    # Close the file...
    f.close()


def tweet_iterator(filename):
    for line in line_iterator(filename):
        yield json.loads(line)


TEXT = os.environ.get("TEXT", 'text')
KLASS = os.environ.get("KLASS", 'klass')
VALUE = os.environ.get("VALUE", 'value')


def read_data_labels(filename, get_tweet=TEXT,
                     get_klass=KLASS, maxitems=1e100):
    data, labels = [], []
    count = 0
    for tweet in tweet_iterator(filename):
        count += 1
        try:
            x = get_tweet(tweet) if callable(get_tweet) else tweet[get_tweet]
            y = get_klass(tweet) if callable(get_klass) else tweet[get_klass]
            data.append(x)
            labels.append(str(y))
            if count == maxitems:
                break
        except KeyError as e:
            raise e

    return data, labels


def read_data_values(filename, get_tweet=TEXT, get_value=VALUE, maxitems=1e100):
    data, values = [], []
    count = 0
    for tweet in tweet_iterator(filename):
        count += 1
        try:
            x = get_tweet(tweet) if callable(get_tweet) else tweet[get_tweet]
            y = get_value(tweet) if callable(get_value) else tweet[get_value]
            data.append(x)
            values.append(float(y))
            if count == maxitems:
                break
        except KeyError as e:
            raise e

    return data, values


def read_data(filename, get_tweet=TEXT, maxitems=1e100):
    data = []
    count = 0
    for tweet in tweet_iterator(filename):
        count += 1
        x = get_tweet(tweet) if callable(get_tweet) else tweet[get_tweet]
        data.append(x)
        if count == maxitems:
            break

    return data


def get_class(m):
    """Import class from string

    :param m: string or class to be imported
    :type m: str or class
    :rtype: class

    >>> get_class('microtc.textmodel.TextModel')
    <class 'microtc.textmodel.TextModel'>
    """
    import importlib

    if isinstance(m, str):
        a = m.split('.')
        p = importlib.import_module('.'.join(a[:-1]))
        return getattr(p, a[-1])
    return m


def load_model(fname):
    """Read model from file. The model must be stored using gzip and pickle

    :param fname: filename
    :type fname: str (path)
    """
    import gzip
    import pickle
    with gzip.open(fname, 'r') as fpt:
        _ = pickle.load(fpt)
    return _


def save_model(obj, fname):
    """Store model from file. The model is stored using gzip and pickle

    :param obj: object to store
    :type obj: object
    :param fname: filename
    :type fname: str (path)
    """
    
    import gzip
    import pickle
    with gzip.open(fname, 'w') as fpt:
        pickle.dump(obj, fpt)


class Counter(collections.Counter):
    def __init__(self, iter=None, update_calls=0, **kwargs):
        super().__init__(iter)
        self.update_calls = update_calls

    @property
    def update_calls(self):
        """
        Count the number of times :py:func:`update` has been called.
        """
        try:
            return self._update_calls
        except AttributeError:
            self._update_calls = -1
        return self._update_calls

    @update_calls.setter
    def update_calls(self, value):
        self._update_calls = value
    
    def update(self, *args, **kwargs):
        self.update_calls += 1
        return super().update(*args, **kwargs)

    def __reduce__(self):
        _ = dict(self)
        return self.__class__, (_, self.update_calls, )



# Copyright 2016-2017 Eric S. Tellez

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import re
import os

from collections import defaultdict
from .params import OPTION_DELETE, OPTION_GROUP, OPTION_NONE


# def get_compiled_map(filename):
#     with open(filename) as f:
#         E = json.load(f)

#     X = defaultdict(list)

#     for code, klass in E.items():
#         X[klass].append(re.escape(code))
            
#     Y = {}

#     for klass, codelist in X.items():
#         Y[klass] = re.compile(r"\b{0}\b".format("|".join(codelist)), re.IGNORECASE)
        
#     return Y


# def transform_replace_by_klass(text, map):
#     for klass, regex in map.items():
#         text = regex.sub(" {0} ".format(klass), text)

#     return re.sub(r"\s+", " ", text)


# def transform_del(text, map):
#     for klass, regex in map.items():
#         text = regex.sub(' ', text)

#     return re.sub(r"\s+", " ", text).strip()


class EmoticonClassifier:
    def __init__(self, fname=None):
        if fname is None:
            fname = os.path.join(os.path.dirname(__file__), 'resources', 'emoticons.json')

        self.emolen = defaultdict(dict)
        self.emoreg = []
        self.some = {}

        with open(fname) as f:
            X = json.load(f)

        for c, k in X.items():  # code, klass
            if c.isalpha():
                r = re.compile(r"\b{0}\b".format(c), re.IGNORECASE)
                self.emoreg.append((r, k))
            else:
                self.emolen[len(c)].setdefault(c, k)

            self.some[c[0]] = max(len(c), self.some.get(c[0], 0))

        maxlen = max(self.emolen.keys())
        self.emolen = [self.emolen.get(i, {}) for i in range(maxlen+1)]

    def replace(self, text, option=OPTION_GROUP):
        if option == OPTION_NONE:
            return text

        for pat, klass in self.emoreg:
            if option == OPTION_DELETE:
                klass = ''
 
            text = pat.sub(klass, text)

        T = []
        i = 0
        _text = text.lower()
        while i < len(text):
            replaced = False
            if _text[i] in self.some:
                for lcode in range(1, len(self.emolen)):
                    if i + lcode < len(_text):
                        code = _text[i:i+lcode]
                        klass = self.emolen[lcode].get(code, None)

                        if klass:
                            if option == OPTION_DELETE:
                                klass = ''

                            T.append(klass)
                            replaced = True
                            i += lcode
                            break
            
            if not replaced:
                T.append(text[i])
                i += 1

        return "".join(T)


def convert_emoji(emoji):
    """Convert the code points into characters

    :param emoji: code point
    :return: emoji
    :rtype: str
    """
    if emoji.count(".."):
        init, end = [int(x, base=16) for x in emoji.split("..")]
        return [chr(x) for x in range(init, end + 1)]
    emojis = emoji.split()
    if len(emojis) > 1:
        emojis = filter(len, [x.strip() for x in emojis])
        return "".join([convert_emoji(x) for x in emojis])
    return chr(int(emoji, base=16))

    
def read_emoji_standard(fname, emos=None):
    """Read the emoji standard files

    :param fname: Path to the file
    :type fname: str
    :param emos: Dictionary computed from a previous called
    :type emos: dict
    :return: Emoji with types as dictionary 
    :rtype: dict
    """

    emos = dict() if emos is None else emos
    with open(fname, encoding="utf-8") as fpt:
        for line in fpt.readlines():
            try:
                line = line[:line.index("#")].strip()            
                if len(line) == 0:
                    continue                
                value, tipo = [x.strip() for x in line.split(";")][:2]
            except ValueError:
                continue
            value = convert_emoji(value)
            values = value if isinstance(value, list) else [value]
            for value in values:
                lst = emos.get(value, list())
                lst.append(tipo)
                emos[value] = lst
    return emos


def create_data_structure(emojis):
    """Create data structure to store the emojis
    :param emojis: Dictionary of emoji
    :type emojis: dict
    :rtype: dict
    """

    head = dict()
    for word in emojis.keys():
        current = head
        for char in word:
            try:
                current = current[char]
            except KeyError:
                _ = dict()
                current[char] = _
                current = _
        current["__end__"] = True
    return head


def find_emoji(data, text):
    """Test whether text has an emoji

    :param data: Output of :py:func:`create_data_structure`
    :type data: dict
    :return: list of pairs, init and end of the word
    :rtype: list
    """

    blocks = list()
    init = i = end = 0
    head = data
    current = head
    while i < len(text):
        char = text[i]
        try:
            current = current[char]
            i += 1
            if "__end__" in current:
                end = i
        except KeyError:
            current = head
            if end > init:
                blocks.append([init, end])
                init = i = end
            elif i > init:
                init = end = i
            else:
                init += 1
                i = end = init
    if end > init:
        blocks.append([init, end])                
    return blocks



# Copyright 2016-2017 Eric S. Tellez

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
import unicodedata
import numpy as np
from .params import OPTION_DELETE, OPTION_GROUP, OPTION_NONE
from .emoticons import EmoticonClassifier
import os
from scipy.sparse import csr_matrix
from .utils import get_class


PUNCTUACTION = ";:,.@\\-\"'/"
SYMBOLS = "()[]¿?¡!{}~<>|"
SKIP_SYMBOLS = set(";:,.@\\-\"/" + SYMBOLS)
SKIP_SYMBOLS_AND_SPACES = set(";:,.@\\-\"/" + SYMBOLS + '\t\n\r ')
# SKIP_WORDS = set(["…", "..", "...", "...."])
WEIGHTING = dict(tfidf="microtc.weighting.TFIDF",
                 tf="microtc.weighting.TFIDF",
                 entropy="microtc.weighting.Entropy")


def norm_chars(text, del_diac=True, del_dup=True, del_punc=False):
    """
    Transform text by removing diacritics, duplicates, and punctuation.
    It adds ~ at the beginning, the end, and the spaces are changed by ~.

    :param text: Text
    :type text: str
    :param del_diac: Delete diacritics
    :type del_diac: bool
    :param del_dup: Delete duplicates
    :type del_dup: bool
    :param del_punc: Delete punctuation symbols
    :type del_punc: bool
    :rtype: str

    Example:

    >>> from microtc.textmodel import norm_chars
    >>> norm_chars("Life is good at Méxicoo.")
    '~Life~is~god~at~Mexico.~'

    """
    L = ['~']

    prev = '~'
    for u in unicodedata.normalize('NFD', text):
        if del_diac:
            o = ord(u)
            if 0x300 <= o and o <= 0x036F:
                continue
            
        if u in ('\n', '\r', ' ', '\t'):
            u = '~'
        elif del_dup and prev == u:
            continue
        elif del_punc and u in SKIP_SYMBOLS:
            prev = u
            continue

        prev = u
        L.append(u)

    L.append('~')

    return "".join(L)


def get_word_list(text):
    """
    Transform a text (begining and ending with ~) to list words.
    It is called after :py:func:`microtc.textmodel.norm_chars`.

    Example

    >>> from microtc.textmodel import get_word_list
    >>> get_word_list("~Someone's house.~")
    ["Someone's", 'house']

    :param text: text
    :type text: str

    :rtype: list
    """

    L = []
    prev = ' '
    for u in text[1:len(text)-1]:
        if u in SKIP_SYMBOLS:
            u = ' '

        if prev == ' ' and u == ' ':
            continue

        if prev == ' ' and u == "'":
            continue

        L.append(u)
        prev = u

    return ("".join(L)).split()


def expand_qgrams(text, qsize, output):
    """Expands a text into a set of q-grams

    :param text: Text
    :type text: str
    :param qsize: q-gram size
    :type qsize: int
    :param output: output
    :type output: list

    :returns: output
    :rtype: list

    Example:

    >>> from microtc.textmodel import expand_qgrams
    >>> output = list()
    >>> expand_qgrams("Good morning.", 3, output)
    ['q:Goo', 'q:ood', 'q:od ', 'q:d m', 'q: mo', 'q:mor', 'q:orn', 'q:rni', 'q:nin', 'q:ing', 'q:ng.']
    """

    _ = ["".join(a) for a in zip(*[text[i:] for i in range(qsize)])]
    [output.append("q:" + x) for x in _]
    return output


def expand_qgrams_word_list(wlist, qsize, output, sep='~'):
    """Expands a list of words into a list of q-grams. It uses `sep` to join words

    :param wlist: List of words computed by :py:func:`microtc.textmodel.get_word_list`.
    :type wlist: list
    :param qsize: q-gram size of words
    :type qsize: int
    :param output: output
    :type output: list
    :param sep: String used to join the words
    :type sep: str

    :returns: output
    :rtype: list

    Example:

    >>> from microtc.textmodel import expand_qgrams_word_list
    >>> wlist = ["Good", "morning", "Mexico"]
    >>> expand_qgrams_word_list(wlist, 2, list())
    ['Good~morning', 'morning~Mexico']
    """

    n = len(wlist)

    for start in range(n - qsize + 1):
        t = sep.join(wlist[start:start+qsize])
        output.append(t)

    return output


def expand_skipgrams_word_list(wlist, qsize, output, sep='~'):
    """Expands a list of words into a list of skipgrams. It uses `sep` to join words

    :param wlist: List of words computed by :py:func:`microtc.textmodel.get_word_list`.
    :type wlist: list
    :param qsize: (qsize, skip) qsize is the q-gram size and skip is the number of words ahead.
    :type qsize: tuple
    :param output: output
    :type output: list
    :param sep: String used to join the words
    :type sep: str

    :returns: output
    :rtype: list

    Example:

    >>> from microtc.textmodel import expand_skipgrams_word_list
    >>> wlist = ["Good", "morning", "Mexico"]
    >>> expand_skipgrams_word_list(wlist, (2, 1), list())
    ['Good~Mexico']

    """
    n = len(wlist)
    qsize, skip = qsize
    for start in range(n - (qsize + (qsize - 1) * skip) + 1):
        if qsize == 2:
            t = wlist[start] + sep + wlist[start+1+skip]
        else:
            t = sep.join([wlist[start + i * (1+skip)] for i in range(qsize)])

        output.append(t)

    return output


class TextModel:
    """

    :param docs: Corpus
    :type docs: list
    :param text: In the case corpus is a dict then text is the key containing the text
    :type text: str
    :param num_option: Transformations on numbers (none | group | delete)
    :type num_option: str
    :param usr_option: Transformations on users (none | group | delete)
    :type usr_option: str
    :param url_option: Transformations on urls (none | group | delete)
    :type url_option: str
    :param emo_option: Transformations on emojis and emoticons (none | group | delete)
    :type emo_option: str
    :param hashtag_option: Transformations on hashtag (none | group | delete)
    :type hashtag_option: str
    :param ent_option: Transformations on entities (none | group | delete)
    :type ent_option: str

    :param lc: Lower case
    :type lc: bool
    :param del_dup: Remove duplicates e.g. hooola -> hola
    :type del_dup: bool
    :param del_punc: Remove punctuation symbols
    :type del_punc: True
    :param del_diac: Remove diacritics
    :type del_diac: bool
    :param token_list: Tokens > 0 qgrams < 0 word-grams
    :type token_list: list
    :param token_min_filter: Keep those tokens that appear more times than the parameter (used in weighting class)
    :type token_min_filter: int or float
    :param token_max_filter: Keep those tokens that appear less times than the parameter (used in weighting class)
    :type token_max_filter: int or float

    :param select_ent:
    :type select_ent: bool
    :param select_suff:
    :type select_suff: bool
    :param select_conn:
    :type select_conn: bool

    :param weighting: Weighting scheme (tfidf | tf | entropy)
    :type weighting: class or str

    Usage:

    >>> from microtc.textmodel import TextModel
    >>> corpus = ['buenos dias', 'catedras conacyt', 'categorizacion de texto ingeotec']

    Using default parameters

    >>> textmodel = TextModel().fit(corpus)

    Represent a text whose words are in the corpus and one that does not

    >>> vector = textmodel['categorizacion ingoetec']
    >>> vector2 = textmodel['cat']

    Using a different token_list

    >>> textmodel = TextModel(token_list=[[2, 1], -1, 3, 4]).fit(corpus)
    >>> vector = textmodel['categorizacion ingoetec']
    >>> vector2 = textmodel['cat']

    Train a classifier

    >>> from sklearn.svm import LinearSVC
    >>> y = [1, 0, 0]
    >>> textmodel = TextModel().fit(corpus)
    >>> m = LinearSVC().fit(textmodel.transform(corpus), y)
    >>> m.predict(textmodel.transform(corpus))
    array([1, 0, 0])
    """

    def __init__(self, docs=None, text='text', num_option=OPTION_GROUP,
                 usr_option=OPTION_GROUP, url_option=OPTION_GROUP,
                 emo_option=OPTION_GROUP, hashtag_option=OPTION_NONE,
                 ent_option=OPTION_NONE, lc=True, del_dup=True, del_punc=False, del_diac=True,
                 token_list=[-1], token_min_filter=0,
                 token_max_filter=1, select_ent=False, select_suff=False, select_conn=False,
                 weighting='tfidf'):
        self._text = os.getenv('TEXT', default=text)
        self.del_diac = del_diac
        self.num_option = num_option
        self.usr_option = usr_option
        self.url_option = url_option
        self.emo_option = emo_option
        self.ent_option = ent_option
        self.select_ent = select_ent
        self.select_suff = select_suff
        self.select_conn = select_conn

        self.hashtag_option = hashtag_option
        self.lc = lc
        self.del_dup = del_dup
        self.del_punc = del_punc
        self.token_list = token_list
        self.token_min_filter = token_min_filter
        self.token_max_filter = token_max_filter
        self.weighting = weighting
        self.weighting = WEIGHTING.get(weighting, weighting)

        if emo_option == OPTION_NONE:
            self.emo_map = None
        else:
            self.emo_map = EmoticonClassifier()

        if docs is not None and len(docs):
            self.fit(docs)

    def fit(self, X):
        """
        Train the model

        :param X: Corpus
        :type X: list
        :rtype: instance
        """

        tokens = [self.tokenize(d) for d in X]
        self.model = get_class(self.weighting)(tokens, X=X,
                                               token_min_filter=self.token_min_filter,
                                               token_max_filter=self.token_max_filter)
        return self

    def __getitem__(self, text):
        """Convert text into a vector

        :param text: Text to be transformed
        :type text: str

        :rtype: list
        """
        return self.model[self.tokenize(text)]

    @classmethod
    def params(cls):
        """
        Parameters

        >>> from microtc.textmodel import TextModel
        >>> TextModel.params()
        odict_keys(['docs', 'text', 'num_option', 'usr_option', 'url_option', 'emo_option', 'hashtag_option', 'ent_option', 'lc', 'del_dup', 'del_punc', 'del_diac', 'token_list', 'token_min_filter', 'token_max_filter', 'select_ent', 'select_suff', 'select_conn', 'weighting'])
        """

        import inspect
        sig = inspect.signature(cls)
        params = sig.parameters.keys()
        return params

    def transform(self, texts):
        """Convert test into a vector

        :param texts: List of text to be transformed
        :type texts: list

        :rtype: list

        Example:

        >>> from microtc.textmodel import TextModel
        >>> corpus = ['buenos dias catedras', 'catedras conacyt']
        >>> textmodel = TextModel().fit(corpus)
        >>> X = textmodel.transform(corpus)
        """
        return self.tonp([self.__getitem__(x) for x in texts])

    def vectorize(self, text):
        raise RuntimeError('Not implemented')

    def tokenize(self, text):
        """Transform text to tokens.
        The procedure is:

        - :py:func:`microtc.textmodel.TextModel.text_transformations`.
        - :py:func:`microtc.textmodel.TextModel.compute_tokens`.
        - :py:func:`microtc.textmodel.TextModel.select_tokens`.

        :param text: Text
        :type text: str or list

        :rtype: list

        Example:

        >>> from microtc.textmodel import TextModel
        >>> tm = TextModel()
        >>> tm.tokenize("buenos dias")
        ['buenos', 'dias']
        >>> tm.tokenize(["buenos", "dias", "tenga usted"])
        ['buenos', 'dias', 'tenga', 'usted']
        """

        if isinstance(text, (list, tuple)):
            tokens = []
            for _text in text:
                tokens.extend(self._tokenize(_text))

            return tokens
        else:
            return self._tokenize(text)

    def get_text(self, text):
        """Return self._text key from text

        :param text: Text
        :type text: dict
        """

        return text[self._text]

    def text_transformations(self, text):
        """
        Text transformations. It starts by analyzing emojis, hashtags, entities,
        lower case, numbers, URL, and users. After these transformations are applied
        to the text, it calls :py:func:`microtc.textmodel.norm_chars`.

        :param text:
        :type text: str

        :rtype: str

        Example:

        >>> from microtc.textmodel import TextModel
        >>> tm = TextModel(del_dup=False)
        >>> tm.text_transformations("Life is good at México @mgraffg.")
        '~life~is~good~at~mexico~_usr~'
        """

        if text is None:
            text = ''

        if isinstance(text, dict):
            text = self.get_text(text)

        if self.emo_map:
            text = self.emo_map.replace(text, option=self.emo_option)

        if self.select_ent:
            text = " ".join(re.findall(r"(@\S+|#\S+|[A-Z]\S+)", text))

        if self.hashtag_option == OPTION_DELETE:
            text = re.sub(r"#\S+", "", text)
        elif self.hashtag_option == OPTION_GROUP:
            text = re.sub(r"#\S+", "_htag", text)

        if self.ent_option == OPTION_DELETE:
            text = re.sub(r"[A-Z][a-z]+", "", text)
        elif self.ent_option == OPTION_GROUP:
            text = re.sub(r"[A-Z][a-z]+", "_ent", text)

        if self.lc:
            text = text.lower()

        if self.num_option == OPTION_DELETE:
            text = re.sub(r"\d+\.?\d+", "", text)
        elif self.num_option == OPTION_GROUP:
            text = re.sub(r"\d+\.?\d+", "_num", text)

        if self.url_option == OPTION_DELETE:
            text = re.sub(r"https?://\S+", "", text)
        elif self.url_option == OPTION_GROUP:
            text = re.sub(r"https?://\S+", "_url", text)

        if self.usr_option == OPTION_DELETE:
            text = re.sub(r"@\S+", "", text)
        elif self.usr_option == OPTION_GROUP:
            text = re.sub(r"@\S+", "_usr", text)

        return norm_chars(text, del_diac=self.del_diac, del_dup=self.del_dup,
                          del_punc=self.del_punc)

    def compute_tokens(self, text):
        """
        Compute tokens from a text using q-grams of characters and words, and skip-grams.

        :param text: Text transformed by :py:func:`microtc.textmodel.TextModel.text_transformations`.
        :type text: str

        :rtype: list

        Example:

        >>> from microtc.textmodel import TextModel
        >>> tm = TextModel(token_list=[-2, -1])
        >>> tm.compute_tokens("~Good morning~")
        [['Good~morning'], ['Good', 'morning']]

        """
        L = []
        textlist = None

        _text = text
        for q in self.token_list:
            if isinstance(q, int):
                if q < 0:
                    if textlist is None:
                        textlist = get_word_list(text)
                    _ = list()
                    expand_qgrams_word_list(textlist, abs(q), _)
                    L.append(_)
                else:
                    _ = list()
                    expand_qgrams(_text, q, _)
                    L.append(_)
            else:
                if textlist is None:
                    textlist = get_word_list(text)

                _ = list()
                expand_skipgrams_word_list(textlist, q, _)
                L.append(_)

        return L

    def select_tokens(self, L):
        """
        Filter tokens using suffix or connections

        :param L: list of tokens
        :type L: list

        :rtype: list
        """

        if self.select_suff:
            L = [tok for tok in L if tok[-1] in SKIP_SYMBOLS_AND_SPACES]
            
        if self.select_conn:
            L = [tok for tok in L if '~' in tok and tok[0] != '~' and tok[-1] != '~']
        return L

    def _tokenize(self, text):
        text = self.text_transformations(text)
        L = []
        for _ in self.compute_tokens(text):
            L += _
        L = self.select_tokens(L)
        if len(L) == 0:
            L = ['~']

        return L

    @property
    def num_terms(self):
        """Dimension which is the number of terms of the corpus

        >>> from microtc.textmodel import TextModel
        >>> corpus = ['buenos dias', 'catedras conacyt', 'categorizacion de texto ingeotec']
        >>> textmodel = TextModel().fit(corpus)
        >>> _ = textmodel.transform(corpus)
        >>> textmodel.num_terms
        8

        :rtype: int
        """

        return self.model.num_terms

    def tonp(self, X):
        """Sparse representation to sparce matrix

        :param X: Sparse representation of matrix
        :type X: list
        :rtype: csr_matrix

        Example:

        >>> from microtc.textmodel import TextModel
        >>> tm = TextModel()
        >>> class A: pass
        >>> tm.model = A()
        >>> tm.model.num_terms = 4
        >>> matrix = [[(1, 0.5), (3, -0.2)], [(2, 0.3)], [(0, 1), (3, -1.2)]]
        >>> r = tm.tonp(matrix)
        >>> r.toarray()
        array([[ 0. ,  0.5,  0. , -0.2],
               [ 0. ,  0. ,  0.3,  0. ],
               [ 1. ,  0. ,  0. , -1.2]])

        """

        if not isinstance(X, list):
            return X
        data = []
        row = []
        col = []
        for r, x in enumerate(X):
            cc = [_[0] for _ in x if np.isfinite(_[1]) and (self.num_terms is None or _[0] < self.num_terms)]
            col += cc
            data += [_[1] for _ in x if np.isfinite(_[1]) and (self.num_terms is None or _[0] < self.num_terms)]
            _ = [r] * len(cc)
            row += _
        return csr_matrix((data, (row, col)), shape=(len(X), self.num_terms))



# Copyright 2016 Ranyart R. Suarez (https://github.com/RanyartRodrigo) and Mario Graff (https://github.com/mgraffg)
# with collaborations of Eric S. Tellez

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sklearn.svm import LinearSVC, LinearSVR
import numpy as np
from scipy.sparse import csr_matrix


class ClassifierWrapper(object):
    def __init__(self, algo=LinearSVC):
        # from sklearn.neighbors.nearest_centroid import NearestCentroid
        # self.svc = NearestCentroid(metric='cosine', shrink_threshold=None)
        # from sklearn.neighbors import KNeighborsClassifier
        # self.svc = KNeighborsClassifier(3, metric='cosine')
        self.svc = algo()

    @property
    def num_terms(self):
        """Dimension which is the number of terms of the corpus

        :rtype: int
        """

        try:
            return self._num_terms
        except AttributeError:
            self._num_terms = None
        return None

    def tonp(self, X):
        """Sparse representation to sparce matrix

        :param X: Sparse representation of matrix
        :type X: list
        :rtype: csr_matrix
        """

        data = []
        row = []
        col = []
        for r, x in enumerate(X):
            cc = [_[0] for _ in x if np.isfinite(_[1]) and (self.num_terms is None or _[0] < self.num_terms)]
            col += cc
            data += [_[1] for _ in x if np.isfinite(_[1]) and (self.num_terms is None or _[0] < self.num_terms)]
            _ = [r] * len(cc)
            row += _
        if self.num_terms is None:
            _ = csr_matrix((data, (row, col)))
            self._num_terms = _.shape[1]
            return _
        return csr_matrix((data, (row, col)), shape=(len(X), self.num_terms))

    def fit(self, X, y):
        X = self.tonp(X)
        self.svc.fit(X, y)
        return self

    def decision_function(self, Xnew):
        Xnew = self.tonp(Xnew)
        return self.svc.decision_function(Xnew)

    def predict_proba(self, Xnew):
        Xnew = self.tonp(Xnew)
        return self.svc.predict_proba(Xnew)

    def predict(self, Xnew):
        Xnew = self.tonp(Xnew)
        ynew = self.svc.predict(Xnew)
        return ynew


class RegressorWrapper(ClassifierWrapper):
    def __init__(self, algo=LinearSVR):
        super(RegressorWrapper, self).__init__(algo=algo)



# Copyright 2016 Mario Graff (https://github.com/mgraffg) and Ranyart R. Suarez (https://github.com/RanyartRodrigo)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def test_predict_from_file():
    from microtc.wrappers import ClassifierWrapper
    from microtc.textmodel import TextModel
    from microtc.utils import read_data_labels
    from sklearn.preprocessing import LabelEncoder

    import os
    fname = os.path.dirname(__file__) + '/text.json'
    corpus, labels = read_data_labels(fname)
    t = TextModel(corpus)
    le = LabelEncoder()
    le.fit(labels)
    y = le.transform(labels)
    c = ClassifierWrapper()
    X = [t[x] for x in corpus]
    c.fit(X, y)
    hy = le.inverse_transform(c.predict(X))
    for i in hy:
        assert i in ['POS', 'NEU', 'NEG']




# Copyright 2016 Mario Graff (https://github.com/mgraffg)
# Copyright 2016 Eric S. Tellez <eric.tellez@infotec.mx>

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from time import time
from sklearn.metrics import r2_score
from scipy.stats import pearsonr, spearmanr
from sklearn import preprocessing
from sklearn import model_selection
from microtc.textmodel import TextModel
from microtc.wrappers import RegressorWrapper


class RegressionScoreSampleWrapper(object):
    valid_scores = ['r2', 'pearsonr', 'spearmanr']

    def __init__(self, X, y, Xstatic=[], ystatic=[], ratio=0.8, test_ratio=None, score='r2', classifier=RegressorWrapper, random_state=None):
        assert ratio < 1, "ratio {0} is invalid, valid values are 0 < ratio < 1".format(ratio)
        self.score = score
        self.le = preprocessing.LabelEncoder().fit(y)
        self.create_classifier = classifier
        if test_ratio is None:
            test_ratio = 1.0 - ratio

        I = list(range(len(y)))
        np.random.shuffle(I)
        s = int(np.ceil(len(y) * ratio))
        s_end = int(np.ceil(len(y) * test_ratio))
        y = self.le.transform(y)
        train, test = I[:s], I[s:s+s_end]
        self.train_corpus = [X[i] for i in train]
        self.train_corpus.extend(Xstatic)

        if len(ystatic) > 0:
            ystatic = self.le.transform(ystatic)
            self.train_y = np.hstack((y[train], ystatic))
        else:
            self.train_y = y[train]

        self.test_corpus = [X[i] for i in test]
        self.test_y = y[test]

    def __call__(self, conf_code):
        conf, code = conf_code
        st = time()
        textmodel = TextModel(self.train_corpus, **conf)
        train_X = [textmodel[doc] for doc in self.train_corpus]
        c = self.create_classifier()
        # c.fit(train_X, self.train_y)
        try:
            c.fit(train_X, self.train_y)
        except ValueError:
            conf["_error"] = "this configuration produces an empty matrix"
            conf["_score"] = 0.0
            return conf
    
        test_X = [textmodel[doc] for doc in self.test_corpus]
        pred_y = c.predict(test_X)
        self.compute_score(conf, pred_y)
        conf['_time'] = (time() - st)
        return conf

    def compute_score(self, conf, hy):
        conf['_r2'] = r2_score(self.test_y, hy)
        conf['_spearmanr'] = spearmanr(self.test_y, hy)[0]
        conf['_pearsonr'] = pearsonr(self.test_y, hy)[0]
        conf['_score'] = conf['_' + self.score]
        # print(conf)


class RegressionScoreKFoldWrapper(RegressionScoreSampleWrapper):
    def __init__(self, X, y, Xstatic=[], ystatic=[], nfolds=5, score='r2', classifier=RegressorWrapper, random_state=None):
        self.nfolds = nfolds
        self.score = score
        # self.X = np.array(X)
        self.X = X
        self.Xstatic = Xstatic
        self.le = preprocessing.LabelEncoder().fit(y)
        self.y = self.le.transform(y)
        if len(ystatic) > 0:
            self.ystatic = self.le.transform(ystatic)
        else:
            self.ystatic = []
        self.test_y = self.y
        self.create_classifier = classifier
        self.kfolds = model_selection.KFold(n_splits=nfolds, shuffle=True, random_state=random_state)

    def __call__(self, conf_code):
        conf, code = conf_code
        st = time()
        predY = np.zeros(len(self.y))
        # X = np.array(self.X)
        for train, test in self.kfolds.split(self.X):
            # A = X[train]
            A = [self.X[i] for i in train]
            if len(self.Xstatic) > 0:
                A.extend(self.Xstatic)

            trainY = self.y[train]
            if len(self.ystatic) > 0:
                trainY = np.hstack((trainY, self.ystatic))

            textmodel = TextModel(A, **conf)
            trainX = [textmodel[x] for x in A]

            c = self.create_classifier()
            try:
                c.fit(trainX, trainY)
            except ValueError:
                conf["_error"] = "this configuration produces an empty matrix"
                conf["_score"] = 0.0
                return conf

            testX = [textmodel[self.X[i]] for i in test]
            predY[test] = c.predict(testX)

        self.compute_score(conf, predY)
        conf['_time'] = (time() - st) / self.nfolds
        return conf



# Copyright 2016-2018 Mario Graff (https://github.com/mgraffg)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""":math:`\mu\text{TC}`
==================================

A great variety of text tasks such as topic or spam identification,
user profiling, and sentiment analysis can be posed as a supervised
learning problem and tackled using a text classifier. A text
classifier consists of several subprocesses, some of them are general
enough to be applied to any supervised learning problem, whereas
others are specifically designed to tackle a particular task using
complex and computational expensive processes such as lemmatization,
syntactic analysis, etc. Contrary to traditional approaches,
:math:`\mu\text{TC}` is a minimalist and multi-propose text-classifier able to tackle
tasks independently of domain and language.

The starting point is :py:class:`microtc.textmodel.TextModel`

"""


__version__ = "2.2.2"



# Copyright 2016 Mario Graff (https://github.com/mgraffg)
# with contributions of Eric S. Tellez <eric.tellez@infotec.mx>

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import microtc
import gzip
from microtc.wrappers import ClassifierWrapper, RegressorWrapper
from microtc.utils import read_data, read_data_labels, read_data_values, tweet_iterator
from multiprocessing import cpu_count, Pool
from collections import defaultdict
from .params import ParameterSelection
from .scorewrapper import ScoreKFoldWrapper, ScoreSampleWrapper
from .regscorewrapper import RegressionScoreKFoldWrapper, RegressionScoreSampleWrapper
from .textmodel import TextModel
from .utils import KLASS, VALUE
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import KFold
import numpy as np
import os
import json
from .utils import save_model, load_model


def load_json(filename):
    if filename.endswith(".gz"):
        f = gzip.GzipFile(filename)
        X = json.load(f)
        f.close()
        return X
    else:
        with open(filename) as f:
            return json.load(f)


def balance(X, y):
    C = defaultdict(list)
    for x, label in zip(X, y):
        C[label].append(x)

    size, label = min(map(lambda label: (len(C[label]), label), C.keys()), key=lambda x: x[0])
    for label, xlist in C.items():
        while len(xlist) > size:
            last = xlist.pop()
            i = np.random.randint(0, size)
            if isinstance(xlist[i], list):
                xlist[i] = last
            else:
                xlist[i] = [xlist[i], last]

    X = []
    y = []
    for label, xlist in C.items():
        X.extend(xlist)
        for i in range(len(xlist)):
            y.append(label)

    return X, y


def clean_params(kw):
    params = TextModel.params()
    return {k: v for k, v in kw.items() if k in params}


class CommandLine(object):
    def __init__(self):
        self.parser = argparse.ArgumentParser(description='microtc')
        self.training_set()
        self.predict_kfold()
        self.param_set()
        self.param_search()
        self.version()

    def version(self):
        pa = self.parser.add_argument
        pa('--version',
           action='version', version='microtc %s' % microtc.__version__)

    def predict_kfold(self):
        pa = self.parser.add_argument
        pa('-k', '--kratio', dest='kratio',
           help='Predict the training set using stratified k-fold (k > 1) or a sampling ratio (when 0 < k < 1)',
           default="0.8",
           type=str)

    def training_set(self):
        cdn = 'File containing the training set'
        pa = self.parser.add_argument
        pa('training_set',
           nargs='+',
           default=None,
           help=cdn)

    def param_search(self):
        pa = self.parser.add_argument
        pa('-s', '--sample', dest='samplesize', type=int, default=32, help="The sample size of the parameter")
        pa('-H', '--hillclimbing', dest='hill_climbing', default=False, action='store_true',
            help="Determines if hillclimbing search is also perfomed to improve the selection of parameters")
        pa('-r', '--resume', dest='best_list', default=None, help="Loads the given file and resumes the search process")
        pa('-b', '--balanced', dest='balanced', default=False,
           action='store_true',
           help="Artificially balances the dataset to reduce bias in unbalanced number of examples per class (only works for classification)")
        pa('-n', '--numprocs', dest='numprocs', type=int, default=1, help="Number of processes to compute the best setup")
        pa('-S', '--score', dest='score', type=str, default='macrof1',
           help="The name of the score to be optimized (classification scores: {0}); (regression scores: {1}) it defaults to macrof1".format(
               ScoreSampleWrapper.valid_scores,
               RegressionScoreSampleWrapper.valid_scores
           ))
        pa('--conf', dest='conf', type=str, default=None, help="Do not perform search, just evaluate the given configuration (in json-format)")

    def param_set(self):
        pa = self.parser.add_argument
        pa('-o', '--output-file', dest='output',
           help='File name to store the output')
        pa('--seed', default=0, type=int)

    def get_output(self):
        if self.data.output is None:
            return self.data.training_set[0] + ".output"
        return self.data.output

    def main(self, args=None, params=None):
        self.data = self.parser.parse_args(args=args)
        np.random.seed(self.data.seed)
        if self.data.numprocs == 1:
            pool = None
        elif self.data.numprocs == 0:
            pool = Pool(cpu_count())
        else:
            pool = Pool(self.data.numprocs)

        assert self.data.score.split(":")[0] in ScoreSampleWrapper.valid_scores + RegressionScoreSampleWrapper.valid_scores, "Unknown score {0}".format(self.data.score)
        if self.data.score in RegressionScoreSampleWrapper.valid_scores:
            ScoreSample = RegressionScoreSampleWrapper
            ScoreKFold = RegressionScoreKFoldWrapper
            _read_data = read_data_values
        else:
            ScoreSample = ScoreSampleWrapper
            ScoreKFold = ScoreKFoldWrapper
            _read_data = read_data_labels

        sel = ParameterSelection(params=params)
        X, y = [], []
        Xstatic, ystatic = [], []
        for train in self.data.training_set:
            if train.startswith("static:"):
                X_, y_ = _read_data(train[7:])
                Xstatic.extend([x for x in tweet_iterator(train[7:])])
                ystatic.extend(y_)
            else:
                X_, y_ = _read_data(train)
                X.extend([x for x in tweet_iterator(train)])
                y.extend(y_)

        if self.data.balanced:
            X, y = balance(X, y)

        if ":" in self.data.kratio:
            ratio, test_ratio = self.data.kratio.split(":")
            fun_score = ScoreSample(X, y, Xstatic=Xstatic, ystatic=ystatic, ratio=float(ratio), test_ratio=float(test_ratio), score=self.data.score, random_state=self.data.seed)
        else:
            ratio = float(self.data.kratio)
            if ratio == 1.0:
                raise ValueError('k=1 is undefined')
            if ratio > 1:
                fun_score = ScoreKFold(X, y, Xstatic=Xstatic, ystatic=ystatic, nfolds=int(ratio), score=self.data.score, random_state=self.data.seed)
            else:
                fun_score = ScoreSample(X, y, Xstatic=Xstatic, ystatic=ystatic, ratio=ratio, score=self.data.score, random_state=self.data.seed)

        if self.data.best_list:
            best_list = load_json(self.data.best_list)
        else:
            best_list = None

        if self.data.conf:
            conf = json.loads(self.data.conf)
            best_list = sel.get_best(fun_score, (conf, 'direct-input'))
        else:
            best_list = sel.search(
                fun_score,
                bsize=self.data.samplesize,
                hill_climbing=self.data.hill_climbing,
                pool=pool,
                best_list=best_list
            )

        best_list0 = list(filter(lambda x: '_error' not in x, best_list))
        if len(best_list0) == 0:
            raise Exception("ERROR best_list is empty" + repr(best_list))

        best_list = best_list0
        with open(self.get_output(), 'w') as fpt:
            fpt.write(json.dumps(best_list0, indent=2, sort_keys=True))

        return best_list0


class CommandLineTrain(CommandLine):
    def __init__(self):
        self.parser = argparse.ArgumentParser(description='microtc')
        self.param_set()
        self.training_set()
        self.param_train()
        self.version()

    def param_train(self):
        pa = self.parser.add_argument
        pa('-m', '--model-params', dest='params_fname', type=str, required=True,
           help="TextModel params")
        pa('-i', '--i-th', dest='position', type=int, default=0,  # best by default
           help="i-th model in the set of configurations (defaults to the best, i.e., 0)")
        pa('-l', '--labels', dest='labels', type=str,
           help="a comma separated list of valid labels")
        pa('-b', '--balanced', dest='balanced', default=False, action='store_true',
           help="Artificially balances the dataset to reduce bias in unbalanced number of examples per class")
        pa('--conf', dest='conf', type=str,
           help="Specifies the configuration in JSON-format")
        pa('-R', '--regression', dest='regression', action='store_true',
           help="The model will be a regressor")

    def main(self, args=None):
        self.data = self.parser.parse_args(args=args)
        if self.data.conf:
            best = json.loads(self.data.conf)
        else:
            best = load_json(self.data.params_fname)[self.data.position]
        best = clean_params(best)
        if self.data.regression:
            _read_data = read_data_values
            wrapper = RegressorWrapper
        else:
            _read_data = read_data_labels
            wrapper = ClassifierWrapper

        corpus, values = [], []
        for train in self.data.training_set:
            X_, y_ = _read_data(train)
            corpus.extend([x for x in tweet_iterator(train)])
            values.extend(y_)

        if self.data.balanced:
            corpus, values = balance(corpus, values)

        t = TextModel(corpus, **best)
        if self.data.regression:
            le = None
            y = values
        else:
            le = LabelEncoder()
            if self.data.labels:
                le.fit(self.data.labels.split(','))
            else:
                le.fit(values)

            y = le.transform(values)

        c = wrapper()
        X = [t[x] for x in corpus]
        c.fit(X, y)
        save_model([t, c, le], self.get_output())
        return [t, c, le]


class CommandLinePredict(CommandLine):
    def __init__(self):
        self.parser = argparse.ArgumentParser(description='microtc')
        self.param_set()
        self.training_set()
        self.param_test()
        self.version()

    def param_test(self):
        pa = self.parser.add_argument
        pa('-m', '--model', dest='model', type=str,
           required=True,
           help="SVM Model file name")

    def get_output(self):
        if self.data.output is None:
            return self.data.test_set[0] + ".predicted"

        return self.data.output

    def training_set(self):
        cdn = 'File containing the test set'
        pa = self.parser.add_argument
        pa('test_set',
           default=None,
           help=cdn)
        pa('--ordinal', dest='ordinal', default=None,
           help="rounds a regression prediction to the nearest integer among the given start:end range")

    def main(self, args=None, model_svc_le=None):
        self.data = self.parser.parse_args(args=args)
        if model_svc_le is None:
            model, svc, le = load_model(self.data.model)
        else:
            model, svc, le = model_svc_le

        veclist = []
        for x in read_data(self.data.test_set):
            v = model[x]
            veclist.append(v)

        L = []
        if le is None:
            hy = svc.predict(veclist)

            if self.data.ordinal:
                start, end = self.data.ordinal.split(':')
                start = int(start)
                end = int(end)

                for tweet, pred in zip(tweet_iterator(self.data.test_set), hy):
                    c = round(pred)
                    if c < start:
                        c = start
                    elif c > end:
                        c = end

                    if c == 0:  # handles IEEE's negative cero -0.0
                        c = 0

                    tweet[VALUE] = int(c)
                    L.append(tweet)
            else:
                for tweet, pred in zip(tweet_iterator(self.data.test_set), hy):
                    tweet[VALUE] = pred
                    L.append(tweet)
        else:
            decision_function = None
            predict_proba = None
            try:
                decision_function = svc.decision_function(veclist).tolist()
            except AttributeError:
                try:
                    predict_proba = svc.predict_proba(veclist).tolist()
                except AttributeError:
                    pass

            hyy = le.inverse_transform(svc.predict(veclist))

            for i, tweet in enumerate(tweet_iterator(self.data.test_set)):
                if decision_function is not None:
                    tweet['decision_function'] = decision_function[i]
                if predict_proba is not None:
                    tweet['predict_proba'] = predict_proba[i]

                klass = hyy[i]
                tweet[KLASS] = str(klass)
                tweet['predicted'] = tweet[KLASS]
                L.append(tweet)

        with open(self.get_output(), 'w') as fpt:
            for tweet in L:
                fpt.write(json.dumps(tweet)+"\n")

        return L


class CommandLineTextModel(CommandLinePredict):
    def main(self, args=None):
        self.data = self.parser.parse_args(args=args)
        textmodel, svc, le = load_model(self.data.model)
        L = []
        with open(self.get_output(), 'w') as fpt:
            for tw in tweet_iterator(self.data.test_set):
                tw["vec"] = [[int(a), float(b)] for a, b in textmodel[tw['text']]]
                tw["vecsize"] = svc.num_terms
                L.append(tw)
                fpt.write(json.dumps(tw) + "\n")
        return L


class CommandLineRetrain(CommandLinePredict):
    def __init__(self):
        self.parser = argparse.ArgumentParser(description='microtc')
        pa = self.parser.add_argument
        pa('-m', '--model', default=None, type=str,
           dest='model', help='specifies the input model')
        pa('-o', '--output', default='', type=str, dest='output',
           help='save the modified model to the given output file')
        pa('training_set', default=None, nargs='+', help="The trainset, it can be used to retrain a model or readjust the model")
        pa('-R', '--regression', dest='regression', action='store_true', help="The model will be a regressor")

    def main(self, args=None):
        self.data = self.parser.parse_args(args=args)
        # logging.basicConfig(level=self.data.verbose)
        textmodel, svc, le = load_model(self.data.model)

        if self.data.regression:
            _read_data = read_data_values
            wrapper = RegressorWrapper
        else:
            _read_data = read_data_labels
            wrapper = ClassifierWrapper

        corpus, values = [], []
        for train in self.data.training_set:
            X_, y_ = _read_data(train)
            corpus.extend(X_)
            values.extend(y_)

        if self.data.regression:
            le = None
            y = values
        else:
            y = le.transform(values)

        c = wrapper()
        X = [textmodel[x] for x in corpus]
        c.fit(X, y)
        save_model([textmodel, c, le], self.get_output())
        return [textmodel, c, le]


class CommandLineKfolds(CommandLineTrain):
    def __init__(self):
        super(CommandLineKfolds, self).__init__()
        self.param_kfold()

    def param_kfold(self):
        pa = self.parser.add_argument
        pa('--update-klass', default=False, dest='update_klass',
           action="store_true", help='Indicates whether the klass should be updated (default False)')
        pa('-k', '--kratio', dest='kratio',
           help='Predict the training set using k-fold (k > 1)',
           default="5",
           type=int)

    def main(self, args=None):
        self.data = self.parser.parse_args(args=args)
        assert not self.data.update_klass
        if self.data.conf:
            best = json.loads(self.data.conf)
        else:
            best = load_json(self.data.params_fname)[0]
        best = clean_params(best)
        corpus, labels = [], []
        for train in self.data.training_set:
            X_, y_ = read_data_labels(train)
            corpus.extend([x for x in tweet_iterator(train)])
            labels.extend(y_)
        le = LabelEncoder()
        if self.data.labels:
            le.fit(self.data.labels.split(','))
        else:
            le.fit(labels)
        y = le.transform(labels)
        model_klasses = os.environ.get('TEXTMODEL_KLASSES')

        if model_klasses:
            model_klasses = le.transform(model_klasses.split(','))
            docs_ = []
            labels_ = []
            for i in range(len(corpus)):
                if y[i] in model_klasses:
                    docs_.append(corpus[i])
                    labels_.append(y[i])

            t = TextModel(docs_, **best)
        else:
            t = TextModel(corpus, **best)

        X = [t[x] for x in corpus]
        hy = [None for x in y]
        for tr, ts in KFold(n_splits=self.data.kratio,
                            shuffle=True, random_state=self.data.seed).split(X):
            c = ClassifierWrapper()
            c.fit([X[x] for x in tr], [y[x] for x in tr])
            _ = c.decision_function([X[x] for x in ts])
            [hy.__setitem__(k, v) for k, v in zip(ts, _)]

        i = 0
        with open(self.get_output(), 'w') as fpt:
            for train in self.data.training_set:
                for tweet in tweet_iterator(train):
                    tweet['decision_function'] = hy[i].tolist()
                    i += 1
                    fpt.write(json.dumps(tweet)+"\n")
        return hy


def params(*args, **kwargs):
    c = CommandLine()
    if len(args) == 0:
        args = None
    return c.main(args, params=kwargs)


def train(*args, **kwargs):
    c = CommandLineTrain()
    if len(args) == 0:
        args = None
    return c.main(args, **kwargs)


def retrain(*args, **kwargs):
    c = CommandLineRetrain()
    if len(args) == 0:
        args = None
    return c.main(args, **kwargs)


def predict(*args, **kwargs):
    c = CommandLinePredict()
    if len(args) == 0:
        args = None
    return c.main(args, **kwargs)


def textmodel(*args, **kwargs):
    c = CommandLineTextModel()
    if len(args) == 0:
        args = None
    return c.main(args, **kwargs)


def kfolds(*args, **kwargs):
    c = CommandLineKfolds()
    if len(args) == 0:
        args = None
    return c.main(args, **kwargs)



# Copyright 2016-2018 Eric S. Tellez <eric.tellez@infotec.mx>

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import setup
import microtc

with open('README.rst') as fpt:
    long_desc = fpt.read()


setup(
    name="microtc",
    description="""A generic minimal text classifier""",
    long_description=long_desc,
    version=microtc.__version__,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        'Programming Language :: Python :: 3',
        "Topic :: Scientific/Engineering :: Artificial Intelligence"],

    packages=['microtc', 'microtc/tests', 'microtc/tools'],
    include_package_data=True,
    zip_safe=False,
    package_data={
        'microtc/tests': ['text.json'],
        'microtc/resources': ['emoticons.json'],
    },
    scripts=[
        'microtc/tools/microtc-train',
        'microtc/tools/microtc-retrain',
        'microtc/tools/microtc-predict',
        'microtc/tools/microtc-params',
        'microtc/tools/microtc-textModel',
        'microtc/tools/microtc-perf',
        'microtc/tools/microtc-ensemble',
        'microtc/tools/microtc-kfolds'
    ]
)
